/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#include <errno.h>
#include <vector>
#include <ctime>
#include <cstring>
#include <cstdlib>
#include "METAR.h"
#include "Util.h"
#include "Exception.h"

const char METAR::WxCodes[][WxCodeSize] = {
		/* Intensity or Proximity */
		"-",	/* Light */
		"+",	/* Heavy */
		"VC",	/* In the Vicinity */

		/* Descriptors */
		"MI",	/* Shallow */
		"PR",	/* Partial */
		"BC",	/* Patches */
		"DR",	/* Low Drifting */
		"BL",	/* Blowing */
		"SH",	/* Shower(s) */
		"TS",	/* Thunderstorm */
		"FZ",	/* Freezing */

		/* Precipitation */
		"DZ",	/* Drizzle */
		"RA",	/* Rain */
		"SN",	/* Snow */
		"SG",	/* Snow Grains */
		"IC",	/* Ice Crystals */
		"PE",	/* Ice Pellets */
		"GR",	/* Hail */
		"GS",	/* Small Hail and/or Snow Pellets */
		"UP",	/* Unknown */

		/* Obscuration */
		"BR",	/* Mist */
		"FG",	/* Fog */
		"FU",	/* Smoke */
		"VA",	/* Volcanic Ash */
		"DU",	/* Widespread Dust */
		"SA",	/* Sand */
		"HZ",	/* Hazy */
		"PY",	/* Spray */

		/* Other */
		"PO",	/* Well-Developed Dust/Sand Whirls */
		"SQ",	/* Squalls */
		"FC",	/* Funnel Cloud */
		"SS",	/* Sand/Dust Storm */

		""		/* Empty Terminator */
		};

METAR::METAR()
{
	ValidParseFlag=false;
	Text="";
	Station="";
	ZuluTime=0;
	Visibility=0.0;
	Ceiling=0;
	CeilingCover=CEILINGCOVER_INVALID;
}

METAR::METAR(std::string Text)
{
	ValidParseFlag=false;
	Station="";
	ZuluTime=0;
	Visibility=0.0;
	Ceiling=0;
	CeilingCover=CEILINGCOVER_INVALID;

	Set(Text);
}

METAR::~METAR()
{
}

bool METAR::Set(std::string Text)
{
	this->Text=Text;
	ValidParseFlag=Parse();
	return ValidParseFlag;
}

std::string METAR::GetStation()
{
	return Station;
}

bool METAR::GetValidParseFlag()
{
	return ValidParseFlag;
}

time_t METAR::GetZuluTime()
{
	return ZuluTime;
}

double METAR::GetVisibility()
{
	return Visibility;
}

int METAR::GetCeiling()
{
	return Ceiling;
}

enum METAR::CeilingCover_enum METAR::GetCeilingCover()
{
	return CeilingCover;
}

enum METAR::FR_enum METAR::GetFR()
{
	if (Visibility>5.0 && Ceiling>3000) {
		return FR_VFR;
	} else if (Visibility>=3.0 && Ceiling>=1000) {
		return FR_MVFR;
	} else if (Visibility>=1.0 && Ceiling>=500) {
		return FR_IFR;
	} else {
		return FR_LIFR;
	}
}

/*
 * Example text:
 * KTDZ 101533 AUTO 25011G18KT 10SM OVC033 01/M06 A3016 RMK AO2 SLP218 T00111056 TSNO
 *
 * For all the gory details, the official document for the US appears to be the
 * Federal Meteorological Handbook No. 1, Chapter 12, available here:
 * http://www.weather.gov/oso/oso1/oso12/fmh1/fmh1ch12.htm
 *
 * That document states the following:
 * The coding standards in this chapter conform to WMO Code Forms METAR FM
 * 15-IX Ext. and SPECI FM 16-IX Ext. and the United States Exceptions filed
 * with the WMO. These exceptions reflect national observing practices which
 * differ from practices outlined in the WMO Manual on Codes No. 306.
 *
 * Even this doesn't appear to be definitive, as the NOAA cycle files contain
 * some stations that aren't in the US and so have different fields - CAVOK
 * for example.  I've also seen a CIVOK and even google has no idea what
 * that means.
 *
 * This function only attempts to parse the information used to determine
 * VFR, MVFR, or IFR conditions.
 *
 * Corner cases are handled conservatively.  For example, a visibility
 * given in terms of RVR is simply considered to be zero.
 *
 * The Ceiling value is taken from the lowest ceiling.  The CeilingCover
 * is taken from the highest (and most dense) ceiling.
 *
 * Useful links include:
 * http://adds.aviationweather.gov/metars/description_ifr.php
 * http://adds.aviationweather.gov/metars/description2.php
 *
 */

bool METAR::Parse()
{
	std::vector<std::string> Tokens;
	std::vector<std::string>::iterator Tokeni,Nexti;
	struct tm bdtime;
	int tmp;

	ValidParseFlag=false;
	if (!Text.size())
		return ValidParseFlag;

	Util::Explode(Text," ",&Tokens);
	Tokeni=Tokens.begin();
	if (Tokeni==Tokens.end())
		return ValidParseFlag;

	/* Individual METARs have a METAR or SPECI prefix.  Lines from a cycle
	 * file do not.  Ignore them if they are present.
	 */
	if (Tokeni->compare("METAR")==0 || Tokeni->compare("SPECI")==0) {
		Tokeni++;
		if (Tokeni==Tokens.end())
			return ValidParseFlag;
	}

	/* Station ID */
	Tokeni=Tokens.begin();
	Station=*(Tokeni++);
	if (Tokeni==Tokens.end())
		return ValidParseFlag;

	/* Zulu Time */
	if (Tokeni->size()!=7)
		return ValidParseFlag;

	time_t Now;

	Now=time(NULL);

	try {
		Util::GmTime_r(&bdtime,Now);
	}
	catch (Exception& e) {
		e.Show();
		return false;
	}

	bdtime.tm_hour=atoi(Tokeni->substr(2,2).c_str());
	bdtime.tm_min=atoi(Tokeni->substr(4,2).c_str());

	tmp=atoi(Tokeni->substr(0,2).c_str());
	if (tmp>bdtime.tm_mday) {
		/* The METAR time is at end of last month */
		bdtime.tm_mon--;
	}
	bdtime.tm_mday=tmp;

	ZuluTime=mktime(&bdtime);

	Tokeni++;
	if (Tokeni==Tokens.end())
		return ValidParseFlag;

	/* Observation Type - Automatic, corrected, or not specified for observed */
	if (Tokeni->compare("AUTO")==0 || Tokeni->compare("COR")) {
		Tokeni++;
		if (Tokeni==Tokens.end())
			return ValidParseFlag;
	}

	/* Wind - ignored for now */
	if (Tokeni->size()>2 && Tokeni->substr(Tokeni->size()-2,2).compare("KT")==0) {
		Tokeni++;
		if (Tokeni==Tokens.end())
			return ValidParseFlag;
	}

	if (Tokeni->compare("CAVOK")==0) {
		Visibility=10.0;
		Ceiling=5000;
		CeilingCover=CEILINGCOVER_SKYCLEAR;
		ValidParseFlag=true;
		return ValidParseFlag;
	}

	/* If this doesn't start with an R but contains a V it's a modifier to
	 * the wind indicating its variable direction.  Ignore it.
	 */
	if (Tokeni->substr(0,1).compare("R")!=0 && Tokeni->find_first_of("V")!=std::string::npos) {
		Tokeni++;
		if (Tokeni==Tokens.end())
			return ValidParseFlag;
	}

	/* Visibility */
	Nexti=Tokeni;
	Nexti++;
	if (Nexti==Tokens.end())
		return ValidParseFlag;
	if (Tokeni->size()==1 && Nexti->find_first_of("/")!=std::string::npos && Nexti->substr(Nexti->size()-2,2).compare("SM")==0) {
		/* Tokeni is the first part of a fractional visibility.  These are
		 * encoded with a space between the whole and fractional parts and
		 * are separated by the Explode function.
		 */
		Visibility=atof(Tokeni->c_str());
		Tokeni++;
		if (Tokeni==Tokens.end())
			return ValidParseFlag;
	} else {
		Visibility=0.0;
	}

	if (Tokeni->size()>2 && Tokeni->substr(Tokeni->size()-2,2).compare("SM")==0) {

		if (Tokeni->substr(0,1).compare("M")==0) {
			/* Consider anything that's described as "less than" to simply be zero */
			Visibility+=0.0;
		} else {
			*Tokeni=Tokeni->substr(0,Tokeni->size()-2);	/* Drop the trailing 'SM' */
			if (Tokeni->find_first_of("/")==std::string::npos) {
				/* Whole number */
				Visibility+=atof(Tokeni->c_str());
			} else {
				/* Fraction */
				Visibility+=atof(Tokeni->substr(0,1).c_str())/atof(Tokeni->substr(2,1).c_str());
			}
		}

		Tokeni++;
		if (Tokeni==Tokens.end())
			return ValidParseFlag;
	} else {
		/* No visibility given */
		return ValidParseFlag;
	}

	/* RVR values */
	if (Tokeni->substr(0,1).compare("R")==0 && Tokeni->find_first_of("/")!=std::string::npos) {
		Visibility=0.0;
		Tokeni++;
		if (Tokeni==Tokens.end())
			return ValidParseFlag;
	}

	/* Ignore any present weather fields */
	while (Tokeni!=Tokens.end()) {
		/* If the field is made up of nothing but WxCodes then it's a
		 * present weather field */
		unsigned int ci,wi;	/* Character and WxCode indicies */

		for (ci=0;ci<Tokeni->size();) {
			for (wi=0;strlen(WxCodes[wi])!=0;wi++) {
				if (Tokeni->substr(ci,strlen(WxCodes[wi])).compare(WxCodes[wi])==0)
					break;
			}
			if (strlen(WxCodes[wi])==0) {
				/* No match - this isn't a weather field */
				break;
			} else {
				/* We found a match, keep checking */
				ci+=strlen(WxCodes[wi]);
			}
		}
		if (ci==Tokeni->size()) {
			/* Everything matched, move on to the next */
			Tokeni++;
			if (Tokeni==Tokens.end())
				return ValidParseFlag;
		} else {
			/* Something didn't match - move on to parsing other things */
			break;
		}
	}

	/* Ceiling */
	bool SawValid=false;
	Ceiling=12000;
	CeilingCover=CEILINGCOVER_SKYCLEAR;
	while (Tokeni!=Tokens.end()) {
		if (Tokeni->compare("SKC")==0 || Tokeni->compare("CLR")==0) {
			if (Tokeni->compare("SKC")==0) {
				CeilingCover=CEILINGCOVER_SKYCLEAR;
			} else {
				CeilingCover=CEILINGCOVER_CLEAR;
			}
			break;
		} else if (Tokeni->substr(0,3).compare("FEW")==0 || Tokeni->substr(0,3).compare("SCT")==0) {
			/* Few & Scattered layers are not ceilings */
			if (Tokeni->substr(0,3).compare("FEW")==0) {
				CeilingCover=CEILINGCOVER_FEW;
			} else {
				CeilingCover=CEILINGCOVER_SCATTERED;
			}
			SawValid=true;
			Tokeni++;
		} else if (Tokeni->substr(0,3).compare("BKN")==0) {
			if (Ceiling==12000)
				Ceiling=100*atoi(Tokeni->substr(3,3).c_str());
			CeilingCover=CEILINGCOVER_BROKEN;
			SawValid=true;
			Tokeni++;
		} else if (Tokeni->substr(0,3).compare("OVC")==0) {
			if (Ceiling==12000)
				Ceiling=100*atoi(Tokeni->substr(3,3).c_str());
			CeilingCover=CEILINGCOVER_OVERCAST;
			SawValid=true;
			Tokeni++;
		} else if (Tokeni->substr(0,2).compare("VV")==0) {
			if (Ceiling==12000)
				Ceiling=100*atoi(Tokeni->substr(2,3).c_str());
			CeilingCover=CEILINGCOVER_OBSCURED;
			SawValid=true;
			Tokeni++;
		} else {
			/* We had something valid and now we don't - we must have moved on to other things */
			if (SawValid) {
				break;
			} else {
				/* No ceiling info found */
				Ceiling=0.0;
				CeilingCover=CEILINGCOVER_MISSING;
			}

			Tokeni++;
		}
	}
	/* No ceiling info was found at all */
	if (Tokeni==Tokens.end()  && !SawValid)
		return ValidParseFlag;

	ValidParseFlag=true;
	return ValidParseFlag;
}

std::string METAR::GetText()
{
	return Text;
}

bool operator < (const METAR& left, const METAR& right)
{
	int result;
	/* Sort first by station */
	result=((METAR)left).GetStation().compare(((METAR)right).GetStation());
	if (result<0)
		return true;
	if (result>0)
		return false;
	/* Then sort by timestamp */
	if (((METAR)left).GetZuluTime()<((METAR)right).GetZuluTime())
		return true;

	return false;
}

bool operator == (const METAR& left, const METAR& right)
{
	return ((METAR)left).GetText().compare(((METAR)right).GetText())==0;
}
